/*
 * Copyright 2014 frankmuenster.de
 * 
 * Lizenziert unter der EUPL, Version 1.1 oder - sobald diese von der
 * Europischen Kommission genehmigt wurden - Folgeversionen der EUPL
 * ("Lizenz"); Sie drfen dieses Werk ausschlielich gem dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 * 
 * http://ec.europa.eu/idabc/eupl
 * 
 * Sofern nicht durch anwendbare Rechtsvorschriften gefordert oder in
 * schriftlicher Form vereinbart, wird die unter der Lizenz verbreitete Software
 * "so wie sie ist", OHNE JEGLICHE GEWHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrcklich oder stillschweigend - verbreitet. Die sprachspezifischen
 * Genehmigungen und Beschrnkungen unter der Lizenz sind dem Lizenztext zu
 * entnehmen.
 */

package de.frankmuenster.jameica.qrscanner.gui.dialog.threads;

import java.awt.Dimension;
import java.net.URL;
import java.util.List;
import java.util.Set;
import java.util.concurrent.CopyOnWriteArraySet;
import java.util.concurrent.TimeoutException;

import org.apache.commons.lang.NullArgumentException;
import org.apache.commons.lang.StringUtils;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.widgets.Display;

import com.github.sarxos.webcam.Webcam;
import com.github.sarxos.webcam.WebcamDevice;
import com.github.sarxos.webcam.WebcamException;
import com.github.sarxos.webcam.ds.buildin.WebcamDefaultDevice;
import com.github.sarxos.webcam.ds.buildin.WebcamDefaultDriver;
import com.github.sarxos.webcam.ds.ipcam.IpCamDevice;
import com.github.sarxos.webcam.ds.ipcam.IpCamDeviceRegistry;
import com.github.sarxos.webcam.ds.ipcam.IpCamDriver;
import com.github.sarxos.webcam.ds.ipcam.IpCamMode;

import de.frankmuenster.jameica.qrscanner.QrCodeScanner;
import de.frankmuenster.jameica.qrscanner.WebcamWatchdog;
import de.frankmuenster.jameica.qrscanner.gui.parts.ImageLabelPart;
import de.frankmuenster.jameica.qrscanner.ipcam.DroidCamDevice;
import de.willuhn.jameica.gui.input.TextAreaInput;
import de.willuhn.jameica.system.Application;
import de.willuhn.logging.Level;
import de.willuhn.logging.Logger;
import de.willuhn.util.I18N;

/**
 * Die Klasse erstellt einen Hintergrund-task, der die Default-Webcam ffnet und
 * solange scannt bis ein QR-Code gefunden wurde oder der Task durch
 * {@link QrCodeScannerThread#interrupt()} unterbrochen wird.
 * 
 * @author Frank Mnster
 * @version 1.2
 * 
 */
@SuppressWarnings("PMD.GodClass")
public class QrCodeScannerThread extends Thread {

	/**
	 * Die Listener fr einen QrCodeFound-Event
	 */
	private transient final Set<QrCodeFoundListener> listeners = new CopyOnWriteArraySet<QrCodeFoundListener>();

	/**
	 * i18n Uebersetzer aus unseren Settings
	 */
	private final static I18N I18N = Application.getPluginLoader().getPlugin(QrCodeScanner.class).getResources()
			.getI18N();

	/**
	 * Mindestbreite fr die Webcam-Auflsung
	 */
	private static final int MIN_SIZE_WIDTH = 640;

	/**
	 * Wurde der Thread von auen unterbrochen
	 */
	private transient boolean interrupted;
	/**
	 * TextArea fr Webcam Infos
	 */
	private transient TextAreaInput webcamInfo;
	/**
	 * Das Label fr das Webcambild
	 */
	private transient ImageLabelPart imgLabel;
	/**
	 * Der Name der Webcam, die benutzt werden soll
	 */
	private transient String webcamName;
	/**
	 * Der gefundene QR-Code als String
	 */
	private transient String result;

	/**
	 * Container fr den WebcamInfoText, da dieser in einem anderem Thread
	 * bentigt wird.
	 */
	private transient String webcamInfoText;

	/**
	 * Fgt eine Listener hinzu
	 * 
	 * @param listener
	 *            der hinzugefgt werden soll
	 */
	public final void addListener(final QrCodeFoundListener listener) {
		listeners.add(listener);
	}

	/**
	 * Entfernt einen Listener aus der Liste
	 * 
	 * @param listener
	 *            der entfernt werden soll
	 */
	public final void removeListener(final QrCodeFoundListener listener) {
		listeners.remove(listener);
	}

	/**
	 * Benachrichtigen der registrierten Listeners mit bergabe des gefundenen
	 * Strings
	 */
	private final void notifyListeners() {
		for (final QrCodeFoundListener listener : listeners) {
			listener.notifyOfQrCodeFound(result);
		}
	}

	/**
	 * Erstellt einen neuen QrCodeScanner Thread
	 * 
	 * @param imgLabel
	 *            das ImageLabelPart in, dem das Webcam-Bild dargestellt werden
	 *            soll.
	 * @param webcamInfo
	 *            Textbereich fr WebCam Infos.
	 * @param webcamName
	 *            der Name der Webcam, die benutzt werden soll.
	 * @throws NullArgumentException
	 *             , wenn <b>imgLabel</b> <i>null</i> ist.
	 */
	public QrCodeScannerThread(final ImageLabelPart imgLabel, final TextAreaInput webcamInfo, final String webcamName)
			throws NullArgumentException {

		super();

		if (imgLabel == null) {
			throw new NullArgumentException("imgLabel"); //$NON-NLS-1$
		} else if (webcamInfo == null) {
			throw new NullArgumentException("webcamInfo"); //$NON-NLS-1$
		} else if (webcamName == null) {
			throw new NullArgumentException("webcamName"); //$NON-NLS-1$
		} else {
			this.webcamInfo = webcamInfo;
			this.imgLabel = imgLabel;
			this.webcamName = webcamName;
			this.result = StringUtils.EMPTY;
		}
	}

	/**
	 * 
	 * @see java.lang.Thread#interrupt()
	 */
	@Override
	public void interrupt() {
		interrupted = true;
	}

	/**
	 * 
	 * @see java.lang.Thread#isInterrupted()
	 */
	@Override
	public boolean isInterrupted() {
		return interrupted;
	}

	/**
	 * 
	 * @see java.lang.Thread#run()
	 */
	@Override
	public void run() {

		final Webcam webcam = getWebcamDevice(webcamName);

		if (webcam == null) {
			imgLabel.setText(I18N.tr("Webcam (" + webcamName + ") nicht gefunden!"));
			if (Logger.isLogging(Level.INFO)) {
				Logger.info("Webcam (" + webcamName + ") nicht gefunden!");
			}
		} else {

			// get resolution
			final Dimension[] dim = webcam.getViewSizes();

			int selectDim = 0;
			for (final Dimension dimension : dim) {
				Logger.trace("Size: " + dimension.toString()); //$NON-NLS-1$
				if (dimension.getWidth() >= MIN_SIZE_WIDTH) {
					break;
				}
				selectDim++;
			}
			Logger.debug("Select Size: " + dim[selectDim].toString()); //$NON-NLS-1$
			webcam.setViewSize(dim[selectDim]);

			// start Loop
			if (!isInterrupted()) {
				loop(webcam);
			}
			if (webcam != null) {
				if (webcam.getLock().isLocked())
					webcam.getLock().unlock();
				if (webcam.isOpen())
					webcam.close();
			}
		}

		Logger.debug("Lade Composite Treiber");
		Webcam.resetDriver();
		Webcam.setDriver(WebcamWatchdog.getDriver());

	}

	private void loop(final Webcam webcam) {

		String qrCode = null;

		try {

			// Informiere Benutzer, das Webcam geoeffnet wird
			imgLabel.setText(I18N.tr("ffne Webcam..."));

			if (!webcam.isOpen()) {
				// webcam.dispose();
				webcam.open();
			}
			Logger.debug("Webcam Auflsung: " + webcam.getViewSize().toString());
			// Update Webcam Info

			webcamInfoText = "Test";
			if (webcam != null) {
				if (webcam.getViewSize().width < MIN_SIZE_WIDTH) {
					webcamInfoText = "Achtung: Die Webcam hat eine geringe (" + webcam.getViewSize().width + "x"
							+ webcam.getViewSize().height + ") Auflsung. Kleine QR-Codes werden "
							+ "eventuell nicht erkannt!";
				} else {
					webcamInfoText = "Webcam Auflsung: " + webcam.getViewSize().width + "x"
							+ webcam.getViewSize().height;
				}
			}
			Display.getDefault().asyncExec(new Runnable() {

				/**
				 * 
				 * @see java.lang.Runnable#run()
				 */
				@Override
				public void run() {
					webcamInfo.setValue(webcamInfoText);
				}
			});

			final Point size = new Point(imgLabel.getSize().x, imgLabel.getSize().y);
			final ScanQrImage scanQrImage = new ScanQrImage(webcam, size);

			while (qrCode == null && !isInterrupted()) {

				// Setze Bild
				imgLabel.setImage(scanQrImage.getImage());

				// Checke QRCode
				qrCode = scanQrImage.getQrDecode();

				Thread.sleep(50);
			}

			Logger.debug("Schliesse Webcam...");
			webcam.close();
			Logger.debug(webcam.getLock().getLockFile().getAbsolutePath());
			webcam.getLock().unlock();

			// set result String in case QR-Code was found
			if (qrCode != null) {
				Logger.debug("QR-Code decoded successfully!");
				result = qrCode;
				notifyListeners();
			}

		} catch (Exception e) { // NOPMD by frank on 21.09.16 22:12
			handleExceptions(e);
		}

	}

	private void handleExceptions(final Exception exception) {
		if (exception.getClass() == TimeoutException.class) {
			imgLabel.setText(I18N.tr("Es konnte keine Webcam gefunden werden..."));
			Logger.info("Keine untersttze Webcam gefunden");
		} else if (exception.getClass() == IllegalArgumentException.class) {
			imgLabel.setText(I18N.tr("Auflsung wird nicht untersttzt!"));
			Logger.warn("Auflsung wird von Webcam nicht untersttzt");
		} else if (exception instanceof WebcamException
				&& exception.getMessage().equals("Cannot get initial image from IP camera device DroidCam")) {
			imgLabel.setText(I18N
					.tr("Kann DroidCam nicht starten!\n\nGreift schon eine andere Applikation/Webseite darauf zu?"));
			Logger.warn("Kann DroidCam nicht starten! Evtl. greift schon eine andere Applikation/Webseite darauf zu.");

		} else {
			imgLabel.setText(I18N.tr("Schwerwiegender Fehler"));
			Logger.error("Ein Fehler ist wrend der Webcamabfragen aufgetreten. Details im debug-log");
			Logger.debug("WebcamError: " + exception.toString());
		}
		try {
			Thread.sleep(3000);
		} catch (InterruptedException e1) {
		}
	}

	private Webcam getWebcamDevice(final String name) {

		Webcam webcam = null;
		final WebcamDevice webcamDevice = getDeviceByName(name);

		if (webcamDevice != null) {

			Webcam.resetDriver();
			// USB Cam
			if (webcamDevice instanceof WebcamDefaultDevice) {

				Logger.debug("WebcamDefaultDevice");
				Webcam.setDriver(WebcamDefaultDriver.class);
				webcam = Webcam.getWebcamByName(name);

			} else if (webcamDevice instanceof IpCamDevice) {
				webcam = handleIpWebcam(webcamDevice);
			}
		}

		return webcam;
	}

	private Webcam handleIpWebcam(final WebcamDevice webcamDevice) {

		Webcam webcam;

		// IP Cam
		Logger.debug("IpCamDevice Type: " + webcamDevice.getClass().getSimpleName());
		Webcam.setDriver(IpCamDriver.class);

		if (IpCamDeviceRegistry.isRegistered((IpCamDevice) webcamDevice)) {
			Logger.debug("Unregister " + IpCamDeviceRegistry.unregister(webcamDevice.getName()));
		}

		final URL url = ((IpCamDevice) webcamDevice).getURL();
		final IpCamMode mode = ((IpCamDevice) webcamDevice).getMode();
		final String name = ((IpCamDevice) webcamDevice).getName();
		webcamDevice.dispose();

		IpCamDevice cam;
		String notOnlineText;

		if (webcamDevice instanceof DroidCamDevice) {
			cam = new DroidCamDevice(url);
			notOnlineText = "\nDie DroidCam App unter " + url + " ist nicht online.\n"
					+ "Bitte URL berprfen, bzw DroidCam auf Android Handy starten.\n\n Warte...";
		} else {
			cam = new IpCamDevice(name, url, mode);
			notOnlineText = "\nDie Mini WebCam App ist nicht online.\n"
					+ "Bitte APP auf dem I-Handy neu starten.\n\n Warte...";
		}
		IpCamDeviceRegistry.register(cam);
		webcam = Webcam.getWebcamByName(name);

		boolean isOnline = ((IpCamDevice) webcam.getDevice()).isOnline();
		Logger.debug("Online: " + isOnline);
		// Ist DoidCam gestartet?
		if (!isOnline) {
			imgLabel.setText(I18N.tr(notOnlineText));
		}
		while (!isOnline && !isInterrupted()) {
			isOnline = ((IpCamDevice) webcam.getDevice()).isOnline();
			try {
				Thread.sleep(300);
			} catch (InterruptedException e) {
			}
		}
		return webcam;

	}

	private WebcamDevice getDeviceByName(final String name) {

		WebcamDevice webcamDevice = null;

		final List<WebcamDevice> webcams = Webcam.getDriver().getDevices();
		for (final WebcamDevice webcamDev : webcams) {
			Logger.debug(webcamDev.toString());
			if (webcamDev.getName().equals(name)) {
				webcamDevice = webcamDev;
				break;
			}
		}

		return webcamDevice;
	}
}

/***********************************************************
 * History: $Log QrCodeScannerThread.java.java,v$ $Date:: 2014-04-01 00:56:03$
 * Date of last commit
 * 
 * @Revision 1.2 Date: 08.07.2016 Untersttzung mehrere Webcams hinzugefgt.
 *           Refactoring: Auslagerung von ScanOrImage
 * @Revision 1.1 Date: 31.03.2014 Try-harder zur QR-Code-Erkennung hinzugefgt
 * 
 *************************************************************/
