/*
* Copyright 2014 frankmuenster.de
*
* Lizenziert unter der EUPL, Version 1.1 oder - sobald
* diese von der Europischen Kommission genehmigt wurden -
* Folgeversionen der EUPL ("Lizenz");
* Sie drfen dieses Werk ausschlielich gem
* dieser Lizenz nutzen.
* Eine Kopie der Lizenz finden Sie hier:
*
*   http://ec.europa.eu/idabc/eupl
*
* Sofern nicht durch anwendbare Rechtsvorschriften
* gefordert oder in schriftlicher Form vereinbart, wird
* die unter der Lizenz verbreitete Software "so wie sie
* ist", OHNE JEGLICHE GEWHRLEISTUNG ODER BEDINGUNGEN -
* ausdrcklich oder stillschweigend - verbreitet.
* Die sprachspezifischen Genehmigungen und Beschrnkungen
* unter der Lizenz sind dem Lizenztext zu entnehmen.
*/
package de.frankmuenster.jameica.qrscanner.utils;

import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.image.BufferedImage;
import java.awt.image.ColorModel;
import java.awt.image.DirectColorModel;
import java.awt.image.IndexColorModel;
import java.awt.image.WritableRaster;

import org.eclipse.swt.graphics.ImageData;
import org.eclipse.swt.graphics.PaletteData;
import org.eclipse.swt.graphics.RGB;

/**
 * Utility-Klasse fr Grafikmanipulationen.
 * 
 * @author Frank Mnster
 *
 */
public final class GraphicsUtils {

	/**
	 * Privater Konstruktor fr die Utils-Klasse
	 */
	private GraphicsUtils() {
		throw new AssertionError("Instantiating utility class..."); //$NON-NLS-1$
	}

	/**
	 * Verndert die Gre eines Images
	 * 
	 * @param image
	 *            image das in der Gre verndert werden soll
	 * @param width
	 *            die neue Breite des Images
	 * @param height
	 *            die neue Hhe des Images
	 * @return das in der Gre angepasste Image
	 */
	public static BufferedImage resize(final BufferedImage image, final int width, final int height) {

		final BufferedImage bufferedImage = new BufferedImage(width, height, BufferedImage.TRANSLUCENT);
		final Graphics2D g2d = (Graphics2D) bufferedImage.createGraphics();
		g2d.addRenderingHints(new RenderingHints(RenderingHints.KEY_RENDERING, RenderingHints.VALUE_RENDER_QUALITY));
		g2d.drawImage(image, 0, 0, width, height, null);
		g2d.dispose();

		return bufferedImage;
	}

	/*
	 * The methods convertToAWT and convertToSWT are taken is taken from the
	 * eclipse snippet 156 available under :
	 * http://git.eclipse.org/c/platform/eclipse.platform.swt.git/tree/examples/
	 * org.eclipse.swt.snippets/src/org/eclipse/swt/snippets/Snippet156.java
	 */
	/*******************************************************************************
	 * Copyright (c) 2000, 2004 IBM Corporation and others. All rights reserved.
	 * This program and the accompanying materials are made available under the
	 * terms of the Eclipse Public License v1.0 which accompanies this
	 * distribution, and is available at
	 * http://www.eclipse.org/legal/epl-v10.html
	 *
	 * Contributors: IBM Corporation - initial API and implementation
	 *******************************************************************************/
	/**
	 * Converts a {@link BufferedImage} to a SWT {@link ImageData}.
	 * 
	 * Taken from Eclipse snippet 156. For more information see:
	 * {@link "http://git.eclipse.org/c/platform/eclipse.platform.swt.git/tree/examples/org.eclipse.swt.snippets/src/org/eclipse/swt/snippets/Snippet156.java"}
	 * 
	 * @param bufferedImage
	 *            the AWT {@link BufferedImage}
	 * @return the SWT {@link ImageData}
	 */
	public static ImageData convertToSWT(final BufferedImage bufferedImage) {

		ImageData data = null;

		if (bufferedImage.getColorModel() instanceof DirectColorModel) {
			final DirectColorModel colorModel = (DirectColorModel) bufferedImage.getColorModel();
			final PaletteData palette = new PaletteData(colorModel.getRedMask(), colorModel.getGreenMask(),
					colorModel.getBlueMask());
			data = new ImageData(bufferedImage.getWidth(), bufferedImage.getHeight(), colorModel.getPixelSize(),
					palette);
			for (int y = 0; y < data.height; y++) {
				for (int x = 0; x < data.width; x++) {
					final int rgb = bufferedImage.getRGB(x, y);
					final int pixel = palette.getPixel(new RGB((rgb >> 16) & 0xFF, (rgb >> 8) & 0xFF, rgb & 0xFF)); // NOPMD
					data.setPixel(x, y, pixel);
					// also set the alpha value (ST)
					data.setAlpha(x, y, colorModel.getAlpha(rgb));
				}
			}
			// return data;
		} else if (bufferedImage.getColorModel() instanceof IndexColorModel) {
			final IndexColorModel colorModel = (IndexColorModel) bufferedImage.getColorModel();
			final int size = colorModel.getMapSize();
			final byte[] reds = new byte[size];
			final byte[] greens = new byte[size];
			final byte[] blues = new byte[size];
			colorModel.getReds(reds);
			colorModel.getGreens(greens);
			colorModel.getBlues(blues);
			RGB[] rgbs = new RGB[size];
			for (int i = 0; i < rgbs.length; i++) {
				rgbs[i] = new RGB(reds[i] & 0xFF, greens[i] & 0xFF, blues[i] & 0xFF); // NOPMD
			}
			final PaletteData palette = new PaletteData(rgbs);
			data = new ImageData(bufferedImage.getWidth(), bufferedImage.getHeight(), colorModel.getPixelSize(),
					palette);
			data.transparentPixel = colorModel.getTransparentPixel();
			final WritableRaster raster = bufferedImage.getRaster();
			final int[] pixelArray = new int[1];
			for (int y = 0; y < data.height; y++) {
				for (int x = 0; x < data.width; x++) {
					raster.getPixel(x, y, pixelArray);
					data.setPixel(x, y, pixelArray[0]);
				}
			}
			// return data;
		}
		return data;
	}

	/**
	 * Converts SWT {@link ImageData} into AWT {@link BufferedImage}
	 * 
	 * Taken from Eclipse snippet 156. For more information see:
	 * {@link "http://git.eclipse.org/c/platform/eclipse.platform.swt.git/tree/examples/org.eclipse.swt.snippets/src/org/eclipse/swt/snippets/Snippet156.java"}
	 * 
	 * @param data
	 *            SWT ImageData image to be transfered
	 * @return transfered AWT BurredImage image
	 * 
	 */
	public static BufferedImage convertToAWT(final ImageData data) {

		BufferedImage bufferedImage = null;

		ColorModel colorModel = null;
		final PaletteData palette = data.palette;
		if (palette.isDirect) {
			colorModel = new DirectColorModel(data.depth, palette.redMask, palette.greenMask, palette.blueMask);
			bufferedImage = new BufferedImage(colorModel,
					colorModel.createCompatibleWritableRaster(data.width, data.height), false, null);
			for (int y = 0; y < data.height; y++) {
				for (int x = 0; x < data.width; x++) {
					final int pixel = data.getPixel(x, y);
					final RGB rgb = palette.getRGB(pixel);
					bufferedImage.setRGB(x, y, rgb.red << 16 | rgb.green << 8 | rgb.blue);
				}
			}
			// return bufferedImage;
		} else {
			final RGB[] rgbs = palette.getRGBs();
			byte[] red = new byte[rgbs.length];
			byte[] green = new byte[rgbs.length];
			byte[] blue = new byte[rgbs.length];
			for (int i = 0; i < rgbs.length; i++) {
				final RGB rgb = rgbs[i];
				red[i] = (byte) rgb.red;
				green[i] = (byte) rgb.green;
				blue[i] = (byte) rgb.blue;
			}
			if (data.transparentPixel == -1) {
				colorModel = new IndexColorModel(data.depth, rgbs.length, red, green, blue);
			} else {
				colorModel = new IndexColorModel(data.depth, rgbs.length, red, green, blue, data.transparentPixel);
			}
			bufferedImage = new BufferedImage(colorModel,
					colorModel.createCompatibleWritableRaster(data.width, data.height), false, null);
			final WritableRaster raster = bufferedImage.getRaster();
			int[] pixelArray = new int[1];
			for (int y = 0; y < data.height; y++) {
				for (int x = 0; x < data.width; x++) {
					final int pixel = data.getPixel(x, y);
					pixelArray[0] = pixel;
					raster.setPixel(x, y, pixelArray);
				}
			}
		}
		return bufferedImage;
	}

}
