/*
 * Copyright 2014 frankmuenster.de
 *
 * Lizenziert unter der EUPL, Version 1.1 oder - sobald
 * diese von der Europischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie drfen dieses Werk ausschlielich gem
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 *   http://ec.europa.eu/idabc/eupl
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrcklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschrnkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */

package de.frankmuenster.jameica.qrscanner.utils;

import java.math.BigInteger;
import java.util.HashSet;
import java.util.Locale;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * 
 * Ein Klasse zum Validieren von IBAN und BIC Basiered auf code aus Forum
 * http://www.tutorials.de/java/259520-iban-bic.html
 * 
 * @see <a href=
 *      "www.tutorials.de/java/259520-iban-bic.html">http://www.tutorials.de/java/259520-iban-bic.html</a>
 * @author takidoso / Frank Mnster
 * @version 1.0
 * 
 */
@SuppressWarnings("PMD.CyclomaticComplexity")
public final class IbanAndBicValidator {

	/**
	 * Die ISO-Lnder-Codes
	 */
	final static private Set<String> ISO_COUNTRY_CODES = new HashSet<String>();

	/**
	 * Eingabe ist keine IBAN.
	 */
	final static public int NOT_AN_IBAN = 0;
	/**
	 * Eingabe hat eine falsche Checksum, ist wahrscheinlich nicht gltig.
	 */
	final static public int BAD_IBAN_CHECKNUM = 1;
	/**
	 * Eingabe ist ein gltige IBAN.
	 */
	final static public int GOOD_IBAN = 2;
	/**
	 * Eingabe ist gltige IBAn mit gltiger EU Lnderkennung.
	 */
	final static public int EU_IBAN = 3;

	/**
	 * Eingabe ist keine BIC.
	 */
	final static public int NOT_A_BIC = 0;
	/**
	 * Eingabe entspricht dem BIC Format. Muss aber keine gltige BIC sein.
	 */
	final static public int GOOD_BIC = 1;
	/**
	 * Eingabe entspricht dem BIC Format mit gltiger EU Lnderkennung, muss
	 * aber trotzdem keinen gltige BIC sein.
	 */
	final static public int EU_BIC = 2;

	/**
	 * BigInteger represenation von 97
	 */
	final static private BigInteger BIG_INT_97 = new BigInteger("97");

	/**
	 * Das IBAN Pattern
	 */
	final static private Pattern IBAN_PAT = Pattern.compile("([A-Z]{2})([0-9]{2})([\\w]{1,30})");

	static {
		final String[] cCodes = Locale.getISOCountries();

		for (int i = 0; i < cCodes.length; i++) {
			ISO_COUNTRY_CODES.add(cCodes[i]);
		}
	}

	/**
	 * Privater Konstruktor fr die Utils-Klasse
	 */
	private IbanAndBicValidator() {
		throw new AssertionError("Instantiating utility class..."); //$NON-NLS-1$
	}

	/**
	 * checks if the given IBAN seems to be an iban (without countrycode-check)
	 * and validates the checknumber this routine is a bit cheap but also fast
	 * since no country-codes are checked
	 * 
	 * @param iban
	 *            die zu berprfende IBAN
	 * @return (int) NOT_AN_IBAN , if given String does not matche basic
	 *         IBAN-format BAD_IBAN_CHECKNUM , if basic IBAN-format is OK but
	 *         not the checknumber GOOD_IBAN , if basic IBAN-format and
	 *         hecknumber are OK
	 */
	static public int isIBANCheckNumberOK(final String iban) {
		int returnCode;
		final Matcher matcher = IBAN_PAT.matcher(iban);

		if (matcher.matches()) {
			final StringBuffer ibanBuff = new StringBuffer(34);
			final StringBuffer valueBuff = new StringBuffer(34);

			ibanBuff.append(matcher.group(3)).append(matcher.group(1)).append(matcher.group(2));

			for (int i = 0; i < ibanBuff.length(); i++) {
				int value;
				final char chr = ibanBuff.charAt(i);
				if (chr >= 'A' && chr <= 'Z') {
					value = chr - 'A' + 10;
					valueBuff.append(value);
				} else if (chr >= 'a' && chr <= 'z') {
					value = chr - 'a' + 10;
					valueBuff.append(value);
				} else {
					valueBuff.append(chr);
				}

			}

			final BigInteger val = new BigInteger(valueBuff.toString());

			returnCode = val.divideAndRemainder(BIG_INT_97)[1].equals(BigInteger.ONE) ? GOOD_IBAN : BAD_IBAN_CHECKNUM;

		} else {
			returnCode = NOT_AN_IBAN;
		}

		return returnCode;
	}

	/**
	 * checks if the given IBAN seems to be an iban with countrycode-check and
	 * validates the checknumber if this validation is allright it is checked
	 * whether the country-code is sorts to european regulations payment area or
	 * not
	 * 
	 * the higher the returned value the better the IBAN so if you just want to
	 * know if the IBAN is allright but don't care about the european-area you
	 * just do something like this: <code>
	 *      if (BankingNumberValidation.isIBAN(iban) >= GOOD_IBAN)
	 *      {
	 *              ... do something
	 *      }
	 *  </code>
	 * 
	 * @param iban
	 *            iban die zu berprfende IBAN
	 * @return (int) NOT_AN_IBAN , if given String does not matche basic
	 *         IBAN-format or countrycod is unkown BAD_IBAN_CHECKNUM , if basic
	 *         IBAN-format and countrycode are OK but not the checknumber
	 *         GOOD_IBAN , if basic IBAN-format, countrycode and checknumber are
	 *         OK EU_IBAN , if basic IBAN-format, countrycode, checknumber are
	 *         OK and the countrycode is european
	 * 
	 * @see de.frankmuenster.jameica.qrscanner.utils.EUCountryCodes#isIbanEUCountry(String)
	 */
	static public int isIBAN(final String iban) {
		int returnCode;
		final Matcher matcher = IBAN_PAT.matcher(iban);

		if (matcher.matches() && ISO_COUNTRY_CODES.contains(iban.substring(0, 2))) {
			final StringBuffer ibanBuff = new StringBuffer(34);
			final StringBuffer valueBuff = new StringBuffer(34);

			ibanBuff.append(matcher.group(3)).append(matcher.group(1)).append(matcher.group(2));

			for (int i = 0; i < ibanBuff.length(); i++) {
				int value;
				final char chr = ibanBuff.charAt(i);
				if (chr >= 'A' && chr <= 'Z') {
					value = chr - 'A' + 10;
					valueBuff.append(value);
				} else if (chr >= 'a' && chr <= 'z') {
					value = chr - 'a' + 10;
					valueBuff.append(value);
				} else {
					valueBuff.append(chr);
				}

			}

			final BigInteger val = new BigInteger(valueBuff.toString());

			if (val.divideAndRemainder(BIG_INT_97)[1].equals(BigInteger.ONE)) {
				if (EUCountryCodes.isIbanEUCountry(iban)) {
					returnCode = EU_IBAN;
				} else {
					returnCode = GOOD_IBAN;
				}
			} else {
				returnCode = BAD_IBAN_CHECKNUM;
			}

		} else {
			returnCode = NOT_AN_IBAN;
		}

		return returnCode;
	}

	/**
	 * very simple formal check on given bic checks if the countrycode is
	 * correct or even european assumes that a bic is 11 or 8 characters long
	 * and an that the 4. and 5. offset is the country-code
	 * 
	 * @param bic
	 *            <Sting> der geprueft werden soll.
	 * @return <b>true</b>, falls bic eine gueltige BIC zu sein scheint,
	 *         <b>false</b> sonst.
	 */
	static public int isBIC(final String bic) {
		int returnCode;
		if (bic.length() == 11 || bic.length() == 8) {
			if (EUCountryCodes.isBicEUCountry(bic)) {
				returnCode = EU_BIC;
			} else if (ISO_COUNTRY_CODES.contains(bic)) {
				returnCode = GOOD_BIC;
			} else {
				returnCode = NOT_A_BIC;
			}
		} else {
			returnCode = NOT_A_BIC;
		}

		return returnCode;
	}
}
